@@ -8,7 +8,8 @@ Created on Apr 17, 2020
 """
 
 import argparse
-import os, pickle, sys
+import os
+import pickle
 import pandas as pd
 import warnings
 
@@ -28,29 +29,52 @@ import gc
 # from beautifultable import BeautifulTable
 
 #### Get Version
-__version__ = "unspec"
+FILENAME = pathlib.Path(__file__).stem
+
 try:
     from importlib.metadata import version, PackageNotFoundError
 
-    __version__ = version(__name__)
+    __version__ = version(FILENAME)
 except:
     try:
         # if the python version is <3.8 this module is used as backport
         from importlib_metadata import version, PackageNotFoundError
 
-        __version__ = version(__name__)
+        __version__ = version(FILENAME)
     except:
         # if module or this python package is not installed
         from pkg_resources import get_distribution, DistributionNotFound
 
         try:
-            __version__ = get_distribution(__name__).version
+            __version__ = get_distribution(FILENAME).version
         # except DistributionNotFound:
         except:
             # package is not installed
             __version__ = "unspec"
 ####
 
+#TODO: New function:
+# Instead of a window around a centerprotein a list of proteins can be given.
+# There should be two modes, one where the names of the groups can be given as a list
+# and another where the protein ids can be given as a list with a required reference genome.
+
+#TODO: Seperate the final window annotation from visualization
+
+#TODO: Visualize the tree in the output
+
+#TODO: Implement an optional rerun through the featurefile to identify the found genes and extract other relevant info
+#like intergenic distance, gene annotation, gene symbols and other non-coding genes in between
+
+#TODO: Unify the output widths by putting the genes to be tracked in columns and the gene numbers in between
+
+#TODO: Improve incode documentation
+
+#TODO: Improve usage documentation and give examples or write a tutorial
+
+#TODO: Implement a scoring system that allows the quantify the relative conservedness of the neighborhood
+
+#TODO: Implement a web-version with javascript that can access the entire gff/feature_table of refseq
+
 
 def writeable_dir(prospective_dir):
     """
@@ -76,7 +100,7 @@ def parse_args():
 
     parser._action_groups.pop()
 
-    required = parser.add_argument_group('required arguments')
+    required = parser.add_argument_group("required arguments")
 
     required.add_argument(
         "--tabular-ortholog-groups",
@@ -105,7 +129,8 @@ def parse_args():
     #     help="if option is set, shows intergenic distances of genes surrounding the center gene",
     # )
 
-    n_group = parser.add_argument_group('required arguments (neighborhood)')
+    n_group = parser.add_argument_group("required arguments (neighborhood)")
+    group_ex = n_group.add_mutually_exclusive_group(required=True)
 
     n_group.add_argument(
         "--reference",
@@ -125,15 +150,25 @@ def parse_args():
         help="unique identifier of the central gene of the window",
     )
 
-    n_group.add_argument(
+    group_ex.add_argument(
         "--extension-size",
-        dest="k",
+        dest="ext_size",
         metavar="<int>",
-        type=str,
-        required=True,
+        type=int,
+        required=False,
         help="defines the #features that are co-checked to the left and right of the centerprotein",
     ),
 
+    group_ex.add_argument(
+        "--extension-mask",
+        dest="ext_mask",
+        metavar="<int>",
+        type=int,
+        nargs='+',
+        required=False,
+        help="defines the position of features that are co-checked to the left and right relative to the centerprotein (position 0).",
+    ),
+
     o_group = parser.add_argument_group("optional arguments (output)")
 
     o_group.add_argument(
@@ -173,14 +208,14 @@ def parse_args():
         help="Attempts to replace genome accessions in the outputs with a replacement string. Requires a two-column map file formatted like so: 'genome file accession' <tab> 'replacement string'",
     )
 
-    r_group = parser.add_argument_group('optional arguments (run)')
+    r_group = parser.add_argument_group("optional arguments (run)")
 
     r_group.add_argument(
         "--nprocs",
         dest="nprocs",
         metavar="<int>",
         type=int,
-        default=mp.cpu_count()-1,
+        default=mp.cpu_count() - 1,
         help="Number of CPUs for parallel processing of genomes. Default: Number of CPUs-1",
     )
 
@@ -290,7 +325,7 @@ class Genome:
                     # TODO: Idea: leave the original indices to avoid to problem of messed up
                     # consecutive order of GENE and CDS Features
 
-    def annotWindowOGs(self, window_ogs, ogtable, center_strand, extension_size):
+    def annotWindowOGs(self, window_ogs, ogtable, center_strand, extension_mask):
         def mapPAtoOG(ogset, pa2og_dict):
             if not ogset or (ogset != ogset):
                 return float("nan")
@@ -310,7 +345,9 @@ class Genome:
         ]  # slice OGtable to Genome
         if pruned_ogtable.empty:
             logging.warning(
-                "Pruning Group Table to {} gives empty table. Trying fuzzy name resolving."
+                "[{}] No orthologs found in table. Retrying with fuzzy name resolution.".format(
+                    self.name
+                )
             )
             pruned_ogtable = ogtable[
                 ogtable.index.isin(["_".join(self.name.split("_")[:2])], level=0)
@@ -340,7 +377,7 @@ class Genome:
                     #     pos_dict[og] = [i - extension_size]
                     # else:
                     #     pos_dict[og].append(i - extension_size)
-                    pos_dict[og] = i - extension_size  # relevant position
+                    pos_dict[og] = extension_mask[i]  # relevant position
 
         # reduce dict to relevant ones only
         # reduced_pa2og_dict = {pa:og for pa,og in pa2og_dict.items() if og in window_ogs} #subset of relevant pa2hogs
@@ -360,7 +397,7 @@ class Genome:
 
         try:
             og = ogtable[~ogtable.index.duplicated()].loc[(genome, pa)]["OG"]
-            #Note: default pandas behaviour: multi-index df loc gives a df for dfs with duplicate indexes
+            # Note: default pandas behaviour: multi-index df loc gives a df for dfs with duplicate indexes
         except:
             return -1  # Code for missing protein information in genome
         if not og:
@@ -391,7 +428,7 @@ class Genome:
 
         return indices
 
-    def findProteinAccessionIndicesOfWindow(self, k, center):
+    def findProteinAccessionIndicesOfWindow(self, extension_mask, center):
         """get upstream and downstream neighboring indexes of CDS and their gene feature"""
 
         center_genomic_accession = self.feature_df.loc[center.name]["genomic_accession"]
@@ -405,18 +442,20 @@ class Genome:
         upstream = []
         downstream = []
 
-        for i in range(1, k + 1):
-            if c - i >= 0:
-                upstream.append(df.iloc[c - i].name)
-            else:
-                upstream.append("")
+        for i in extension_mask:
+            if i < 0:
+                if c + i >= 0: #if the contig position exceeds the contig indicate by "" in window
+                    upstream.append(df.iloc[c + i].name)
+                else:
+                    upstream.append("")
 
-            if c + i <= len(df) - 1:
-                downstream.append(df.iloc[c + i].name)
-            else:
-                downstream.append("")
+            if i > 0:
+                if c + i <= len(df) - 1:
+                    downstream.append(df.iloc[c + i].name)
+                else:
+                    downstream.append("")
 
-        return upstream[::-1] + [center.name] + downstream
+        return upstream + [center.name] + downstream
 
     def produceCMDLOutput(self, label_map=None):
 
@@ -649,7 +688,11 @@ def parseGFF3(gff3_path):
                 .str.split(";", n=1, expand=True)[0]
             )
         except:
-            logging.warning("An error occured while parsing the subattribute {} from the gff file. Probably absent.".format(attrid))
+            logging.warning(
+                "An error occured while parsing the subattribute {} from the gff file. Probably absent.".format(
+                    attrid
+                )
+            )
             ft_df[cname] = float("NaN")
 
     ft_df["GeneID"] = ft_df["attributes"].str.extract(r"Dbxref=.*?GeneID:([\d]+)")[0]
@@ -722,11 +765,21 @@ def getTaxonOrder(treepath, ref_path):
     return t.get_leaf_names()
 
 
-def profile_genome(t, window_ogs, ogtable, center, extension_size):
-    if t.suffix in [".gff", ".gff3"]:
-        ft_df = parseGFF3(t)
-    elif t.suffix in [".txt"]:
-        ft_df = parseFeatureTable(t)
+def profile_genome(t, window_ogs, ogtable, center, extension_mask):
+
+    try:
+        if t.suffix in [".gff", ".gff3"]:
+            ft_df = parseGFF3(t)
+        elif t.suffix in [".txt"]:
+            ft_df = parseFeatureTable(t)
+
+    except:
+        logging.error(
+            "[{}] Parsing error. Correct format? Note: '~.txt' files must be of format feature_table, '~.gff' files of format GFF3.".format(
+                t.name
+            )
+        )
+        return None
 
     genome_name = t.stem.replace("_feature_table", "").replace("_genomic", "")
     g = Genome(genome_name, ft_df)
@@ -736,12 +789,10 @@ def profile_genome(t, window_ogs, ogtable, center, extension_size):
     try:
         # only_window_ogtable = ogtable[ogtable["OG"].isin(window_ogs)]
         # g.annotWindowHOGs(window_ogs, only_window_ogtable)
-        g.annotWindowOGs(window_ogs, ogtable, center.loc["strand"], extension_size)
+        g.annotWindowOGs(window_ogs, ogtable, center.loc["strand"], extension_mask)
     except:
-        logging.info(
-            "No orthology information found for the taxon:"
-            + g.name
-            + ". Ignoring taxon."
+        logging.error(
+            "[{}] No orthology information found. Ignoring taxon.".format(t.name)
         )
         tb = traceback.format_exc()
         return None
@@ -750,9 +801,10 @@ def profile_genome(t, window_ogs, ogtable, center, extension_size):
 
     return g
 
-def compare_taxon(t, window_ogs, ogtable, center, extension_size, label_map ):
-    logging.info("[{}] profiling ...".format(t))
-    g = profile_genome(t, window_ogs, ogtable, center, extension_size)
+
+def compare_taxon(t, window_ogs, ogtable, center, extension_mask, label_map):
+    logging.info("[{}] profiling ...".format(t.name))
+    g = profile_genome(t, window_ogs, ogtable, center, extension_mask)
     if g:
         # runprofile.enable()
         str_out = g.produceCMDLOutput(label_map=label_map)
@@ -760,12 +812,12 @@ def compare_taxon(t, window_ogs, ogtable, center, extension_size, label_map ):
         # runprofile.disable()
         # stats = pstats.Stats(runprofile).sort_stats('cumtime')
         # stats.print_stats(10)
-    if str_out:
-        print(str_out)
-        logging.info("[{}] ... OK.".format(t))
-        return str_out
+        if str_out:
+            print(str_out)
+            logging.info("[{}] ... OK.".format(t.name))
+            return str_out
     else:
-        logging.info("[{}] ... Failed.".format(t))
+        logging.warning("[{}] ... Failed. Ignoring taxon.".format(t.name))
     gc.collect()
 
 
@@ -782,9 +834,15 @@ def main():
     if args.output_prefix:
         PREFIX = args.output_prefix
     else:
-        PREFIX = "prot_{}_ext_{}".format(args.centerprotein_accession, args.k)
+        if args.ext_mask:
+            extension_info = "custom_mask"
+        elif args.ext_size:
+            extension_info = args.ext_size
+        PREFIX = "prot_{}_ext_{}".format(args.centerprotein_accession, extension_info)
 
     logfilepath = pathlib.Path(args.outdir) / "{}.vicinator.log".format(PREFIX)
+    if logfilepath.is_file():
+        pathlib.Path.unlink(logfilepath)
     # logFormatter = logging.Formatter("%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
     logging.basicConfig(  # filename=str(logfilepath),
         level=logging.INFO,
@@ -844,9 +902,16 @@ def main():
     ]  # only take the first of all hits of the central protein
 
     # FIND UPSTREAM AND DOWNSTREAM PROTEINS
-    extension_size = int(args.k)
+    if args.ext_size:
+        extension_size = abs(int(args.ext_size))
+        extension_mask = tuple(range(-extension_size, extension_size+1))
+    if args.ext_mask:
+        extension_mask = [int(x) for x in args.ext_mask]
+        if 0 not in extension_mask:
+            extension_mask.append(0)
+        extension_mask = tuple(sorted(extension_mask))
 
-    window = ref_g.findProteinAccessionIndicesOfWindow(extension_size, center)
+    window = ref_g.findProteinAccessionIndicesOfWindow(extension_mask, center)
 
     prot_accessions = [
         ref_g.feature_df.loc[i]["product_accession"] if i != "" else "" for i in window
@@ -858,7 +923,7 @@ def main():
 
     window_ogs = []
     for i, accession_set in enumerate(prot_accessions):
-        #print(i, accession_set)
+        # print(i, accession_set)
         ogs = set()
         for acc in accession_set:
             og = ref_g.getOGidFromTable(ogtable, ref_g.name, acc)
@@ -897,6 +962,7 @@ def main():
         for p in pathlib.Path(args.feat_tables_dir).glob("*")
         if p.is_file() and p.suffix in [".gff", ".gff3", ".txt"]
     ]
+    logging.info("{} feature files (.txt,.gff) found.".format(len(taxon_feature_files)))
     if args.tree:
         taxon_order = getTaxonOrder(args.tree.name, args.ref_feat_table.name)
         ordered_taxon_feature_files = []
@@ -946,7 +1012,13 @@ def main():
     manager = mp.Manager()
     return_dict = manager.dict()
     logging.info("Number of CPUs used: {}".format(args.nprocs))
-    full_output = pool.starmap(compare_taxon, [(t, window_ogs, ogtable, center, extension_size, label_map) for t in taxon_feature_files])
+    full_output = pool.starmap(
+        compare_taxon,
+        [
+            (t, window_ogs, ogtable, center, extension_mask, label_map)
+            for t in taxon_feature_files
+        ],
+    )
 
     # full_output = []
     #
@@ -970,13 +1042,15 @@ def main():
     #         logging.info("[{}] ... Failed.".format(t))
     #     gc.collect()
 
-    full_output = "\n".join(full_output)
+    full_output = "\n".join([f for f in full_output if f])
     htmloutputfilepath = pathlib.Path(args.outdir) / "{}.vicinator.out.html".format(
         PREFIX
     )
     with open(htmloutputfilepath, "w") as htmlout:
         htmlout.write(converter.convert(full_output))
 
+    logging.info("Finished successfully.")
+
 
 if __name__ == "__main__":
     main()
